# !/usr/bin/env python
# coding=utf-8

from __future__ import print_function
from . import tools
from . import decision_tree
import ee
from . import __version__
from .bitreader import BitReader
# options for BitReaders for known collections

# 16 bits
BITS_MODIS09GA = {
    '0-1': {0:'clear', 1:'cloud', 2:'mix'},
    '2':   {1:'shadow'},
    '8-9': {1:'small_cirrus', 2:'average_cirrus', 3:'high_cirrus'},
    '13':  {1:'adjacent'},
    '15':  {1:'snow'}
}

# 16 bits
BITS_MODIS13Q1 = {
    '0-1': {0:'good_qa'},
    '2-5': {0:'highest_qa'},
    '8':   {1:'adjacent'},
    '10':  {1:'cloud'},
    '14':  {1:'snow'},
    '15':  {1:'shadow'}
}

# USGS SURFACE REFLECTANCE
# 8 bits
BITS_LANDSAT_CLOUD_QA = {
    '0': {1:'ddv'},
    '1': {1:'cloud'},
    '2': {1:'shadow'},
    '3': {1:'adjacent'},
    '4': {1:'snow'},
    '5': {1:'water'}
}

# USGS SURFACE REFLECTANCE
# 16 bits
BITS_LANDSAT_PIXEL_QA = {
    '1': {1:'clear'},
    '2': {1:'water'},
    '3': {1:'shadow'},
    '4': {1:'snow'},
    '5': {1:'cloud'},
    '6-7':{3:'high_confidence_cloud'}
}

# USGS SURFACE REFLECTANCE L8
BITS_LANDSAT_PIXEL_QA_L8 = {
    '1': {1:'clear'},
    '2': {1:'water'},
    '3': {1:'shadow'},
    '4': {1:'snow'},
    '5': {1:'cloud'},
    '6-7':{3:'high_confidence_cloud'},
    '8-9':{3:'cirrus'},
    '10': {1:'occlusion'}
}

# USGS TOA
BITS_LANDSAT_BQA = {
    '4': {1:'cloud'},
    '5-6': {3:'high_confidence_cloud'},
    '7-8': {3:'shadow'},
    '9-10': {3:'snow'}
}

# USGS TOA L8
BITS_LANDSAT_BQA_L8 = {
    '4': {1:'cloud'},
    '5-6': {3:'high_confidence_cloud'},
    '7-8': {3:'shadow'},
    '9-10': {3:'snow'},
    '11-12': {3:'cirrus'}
}

# SENTINEL 2
BITS_SENTINEL2 = {
    '10':{1:'cloud'},
    '11':{1:'cirrus'}
}


def decodeBitsEE(bit_reader, qa_band):
    """
    :param bit_reader: the bit reader
    :type bit_reader: BitReader
    :param qa_band: name of the band that holds the bit information
    :type qa_band: str
    :return: a function to map over a collection. The function adds all
        categories masks as new bands
    """
    options = ee.Dictionary(bit_reader.info)
    categories = ee.List(bit_reader.all_categories)

    def wrap(image):
        def eachcat(cat, ini):
            ini = ee.Image(ini)
            qa = ini.select(qa_band)
            # get data for category
            data = ee.Dictionary(options.get(cat))
            lshift = ee.Number(data.get('lshift'))
            length = ee.Number(data.get('bit_length'))
            decoded = ee.Number(data.get('shifted'))
            # move = places to move bits right and left back
            move = lshift.add(length)
            # move bits right and left
            rest = qa.rightShift(move).leftShift(move)
            # subtract the rest
            norest = qa.subtract(rest)
            # right shift to compare with decoded data
            to_compare = norest.rightShift(lshift) ## Image
            # compare if is equal, return 0 if not equal, 1 if equal
            mask = to_compare.eq(decoded)
            # rename to the name of the category
            qa_mask = mask.select([0], [cat])

            return ini.addBands(qa_mask)
        return ee.Image(categories.iterate(eachcat, image))
    return wrap


def generalMask(options, reader, qa_band, update_mask=True,
                add_mask_band=True, add_every_mask=False,
                all_masks_name='mask'):
    """ General function to get a bit mask band given a set of options
    a bit reader and the name of the qa_band

    :param options: options to decode
    :param reader: the bit reader
    :param qa_band: the name of the qa band
    :param updateMask: whether to update the mask for all options or not
    :param addBands: whether to add the mask band for all options or not
    :return: a function to map over a collection
    """
    encoder = decodeBitsEE(reader, qa_band)
    opt = ee.List(options)
    clases = ("'{}', "*len(options))[:-2].format(*options)

    # Property when adding every band
    msg_eb = "Band called 'mask' is for {} and was computed by geetools" \
          " version {} (https://github.com/gee-community/gee_tools)"
    prop_eb = ee.String(msg_eb.format(clases, __version__))
    prop_name_eb = ee.String('{}_band'.format(all_masks_name))

    def add_every_bandF(image, encoded):
        return image.addBands(encoded).set(prop_name_eb, prop_eb)

    def get_all_mask(encoded):
        # TODO: put this function in tools
        initial = encoded.select([ee.String(opt.get(0))])
        rest = ee.List(opt.slice(1))
        def func(cat, ini):
            ini = ee.Image(ini)
            new = encoded.select([cat])
            return ee.Image(ini.Or(new))

        all_masks = ee.Image(rest.iterate(func, initial)) \
            .select([0], [all_masks_name])
        mask = all_masks.Not()
        return mask

    # 0 0 1
    if not add_every_mask and not update_mask and add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            mask = get_all_mask(encoded)
            return image.addBands(mask)

    # 0 1 0
    elif not add_every_mask and update_mask and not add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            mask = get_all_mask(encoded)
            return image.updateMask(mask)

    # 0 1 1
    elif not add_every_mask and update_mask and add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            mask = get_all_mask(encoded)
            return image.updateMask(mask).addBands(mask)

    # 1 0 0
    elif add_every_mask and not update_mask and not add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            return add_every_bandF(image, encoded)

    # 1 0 1
    elif add_every_mask and not update_mask and add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            mask = get_all_mask(encoded)
            return add_every_bandF(image, encoded).addBands(mask)

    # 1 1 0
    elif add_every_mask and update_mask and not add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            mask = get_all_mask(encoded)
            updated = image.updateMask(mask)
            with_bands = add_every_bandF(updated, encoded)
            return with_bands

    # 1 1 1
    elif add_every_mask and update_mask and add_mask_band:
        def wrap(image):
            encoded = encoder(image).select(opt)
            mask = get_all_mask(encoded)
            updated = image.updateMask(mask)
            with_bands = add_every_bandF(updated, encoded)
            return with_bands.addBands(mask)

    return wrap


def modis09ga(options=('cloud', 'mix', 'shadow', 'snow'), update_mask=True,
              add_mask_band=True, add_every_mask=False):
    """ Function for masking MOD09GA and MYD09GA collections

    :return: a function to use in a map function over a collection
    """
    reader = BitReader(BITS_MODIS09GA, 16)
    return generalMask(options, reader, 'state_1km',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def modis13q1(options=('cloud', 'adjacent', 'shadow', 'snow'),
              update_mask=True, add_mask_band=True, add_every_mask=False):
    """ Function for masking MOD13Q1 and MYD13Q1 collections

    :return: a function to use in a map function over a collection
    """
    reader = BitReader(BITS_MODIS13Q1, 16)
    return generalMask(options, reader, 'DetailedQA',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def landsat457SRCloudQA(options=('cloud', 'adjacent', 'shadow', 'snow'),
                        update_mask=True, add_mask_band=True, add_every_mask=False):

    reader = BitReader(BITS_LANDSAT_CLOUD_QA, 8)
    return generalMask(options, reader, 'sr_cloud_qa',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def landsat457SRPixelQA(options=('cloud', 'shadow', 'snow'),
                        update_mask=True, add_mask_band=True, add_every_mask=False):
    reader = BitReader(BITS_LANDSAT_PIXEL_QA, 16)
    return generalMask(options, reader, 'pixel_qa',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def landsat8SRPixelQA(options=('cloud', 'shadow', 'snow', 'cirrus'),
                      update_mask=True, add_mask_band=True, add_every_mask=False):
    reader = BitReader(BITS_LANDSAT_PIXEL_QA_L8, 16)
    return generalMask(options, reader, 'pixel_qa',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def landsat457ToaBQA(options=('cloud', 'shadow', 'snow'),
                     update_mask=True, add_mask_band=True, add_every_mask=False):
    reader = BitReader(BITS_LANDSAT_BQA, 16)
    return generalMask(options, reader, 'BQA',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def landsat8ToaBQA(options=('cloud', 'shadow', 'snow', 'cirrus'),
                   update_mask=True, add_mask_band=True, add_every_mask=False):
    reader = BitReader(BITS_LANDSAT_BQA_L8, 16)
    return generalMask(options, reader, 'BQA',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def sentinel2(options=('cloud', 'cirrus'), update_mask=True,
              add_mask_band=True, add_every_mask=False):
    reader = BitReader(BITS_SENTINEL2, 16)
    return generalMask(options, reader, 'QA60',
                       update_mask=update_mask,
                       add_mask_band=add_mask_band,
                       add_every_mask=add_every_mask)


def compute(image, mask_band, bits, options=None, name_all='all_masks'):
    """ Compute bits using a specified band, a bit's relation and a list of
    options

    :param image: the image that holds the bit mask band
    :type image: ee.Image
    :param mask_band: the name of the band that holds the bits mask
    :type mask_band: str
    :param bits: relation between name and bit
    :type bits: dict
    :param options: list of 'bits' to compute. Example: ['cloud', 'snow']. If
        None, will use all keys of the relation's dict
    :type options: list
    :param name_all: name for the band that holds the final mask. Default:
        'all_masks'
    :type name_all: str
    :return: The computed mask
    :rtype: ee.Image
    """
    # cast params in case they are not EE objects
    bits_dict = ee.Dictionary(bits)
    opt = ee.List(options) if options else bits_dict.keys()
    image = ee.Image(image).select(mask_band)

    first = ee.Image.constant(0).select([0], [name_all]) # init image

    # function for iterate over the options
    def for_iterate(option, ini):
        i = ee.Image(ini) # cast ini
        all = i.select([name_all])

        # bits relation dict contains the option?
        cond = bits_dict.contains(option)

        def for_true():
            """ function to execute if condition == True """
            # get the mask for the option
            mask = tools.image.computeBits(image, bits_dict.get(option),
                                           bits_dict.get(option),
                                           option)

            # name the mask
            # mask = ee.Image(mask).select([0], [option])
            newmask = all.Or(mask)

            # return ee.Image(all.Or(mask)).addBands(mask)
            return tools.image.replace(i, name_all, newmask).addBands(mask)

        return ee.Image(ee.Algorithms.If(cond, for_true(), i))

    good_pix = ee.Image(opt.iterate(for_iterate, first))

    # return good_pix.Not()
    return good_pix


def hollsteinMask(image,
                  options=('cloud', 'snow', 'shadow', 'water', 'cirrus'),
                  aerosol='B1', blue='B2', green='B3', red_edge1='B5',
                  red_edge2='B6', red_edge3='B7', red_edge4='B8A',
                  water_vapor='B9', cirrus='B10', swir='B11',
                  name='hollstein'):
    """ Get Hollstein mask """
    def difference(a, b):
        def wrap(img):
            return img.select(a).subtract(img.select(b))
        return wrap

    def ratio(a, b):
        def wrap(img):
            return img.select(a).divide(img.select(b))
        return wrap

    # 1
    b3 = image.select(green).lt(3190)

    # 2
    b8a = image.select(red_edge4).lt(1660)
    r511 = ratio(red_edge1, swir)(image).lt(4.33)

    # 3
    s1110 = difference(swir, cirrus)(image).lt(2550)
    b3_3 = image.select(green).lt(5250)
    r210 = ratio(blue, cirrus)(image).lt(14.689)
    s37 = difference(green, red_edge3)(image).lt(270)

    # 4
    r15 = ratio(aerosol, red_edge1)(image).lt(1.184)
    s67 = difference(red_edge2, red_edge3)(image).lt(-160)
    b1 = image.select(aerosol).lt(3000)
    r29 =  ratio(blue, water_vapor)(image).lt(0.788)
    s911 = difference(water_vapor, swir)(image).lt(210)
    s911_2 = difference(water_vapor, swir)(image).lt(-970)

    snow = {'snow':[['1',0], ['22',0], ['34',0]]}
    cloud = {'cloud-1':[['1',0], ['22',1],['33',1],['44',1]],
             'cloud-2':[['1',0], ['22',1],['33',0],['45',0]]}
    cirrus = {'cirrus-1':[['1',0], ['22',1],['33',1],['44',0]],
              'cirrus-2':[['1',1], ['21',0],['32',1],['43',0]]}
    shadow = {'shadow-1':[['1',1], ['21',1],['31',1],['41',0]],
              'shadow-2':[['1',1], ['21',1],['31',0],['42',0]],
              'shadow-3':[['1',0], ['22',0],['34',1],['46',0]]}
    water = {'water':[['1',1], ['21',1],['31',0],['42',1]]}

    all = {'cloud':cloud,
           'snow': snow,
           'shadow':shadow,
           'water':water,
           'cirrus':cirrus}

    final = {}

    for option in options:
        final.update(all[option])

    dtf = decision_tree.binary(
        {'1':b3,
         '21':b8a, '22':r511,
         '31':s37, '32':r210, '33':s1110, '34':b3_3,
         '41': s911_2, '42':s911, '43':r29, '44':s67, '45':b1, '46':r15
         }, final, name)

    return dtf


def applyHollstein(image,
                   options=('cloud', 'snow', 'shadow', 'water', 'cirrus'),
                   aerosol='B1', blue='B2', green='B3', red_edge1='B5',
                   red_edge2='B6', red_edge3='B7', red_edge4='B8A',
                   water_vapor='B9', cirrus='B10', swir='B11'):
    """ Apply Hollstein mask """
    mask = hollsteinMask(image, options, aerosol, blue, green, red_edge1,
                         red_edge2, red_edge3, red_edge4, water_vapor,
                         cirrus, swir).select('hollstein')
    return image.updateMask(mask)


def hollsteinS2(options=('cloud', 'snow', 'shadow', 'water', 'cirrus'),
                name='hollstein', addBands=False, updateMask=True):
    """ Compute Hollstein Decision tree for detecting clouds, clouds shadow,
    cirrus, snow and water in Sentinel 2 imagery

    :param options: masks to apply. Options: 'cloud', 'shadow', 'snow',
        'cirrus', 'water'
    :type options: list
    :param name: name of the band that will hold the final mask. Default: 'hollstein'
    :type name: str
    :param addBands: add all bands to the image. Default: False
    :type addBands: bool
    :param updateMask: update the mask of the Image. Default: True
    :type updateMask: bool
    :return: a function for applying the mask
    :rtype: function
    """
    def compute_dt(img):

        results = hollsteinMask(img, options, name)

        if updateMask and addBands:
            return img.addBands(results).updateMask(results.select(name))
        elif addBands:
            return img.addBands(results)
        elif updateMask:
            return img.updateMask(results.select(name))

    return compute_dt


def darkPixels(green, swir2, threshold=0.25):
    """ Detect dark pixels from green and swir2 band

    :param green: name of the green band
    :type green: str
    :param swir2: name of the swir2 band
    :type swir2: str
    :param threshold: threshold value from which are considered dark pixels
    :type threshold: float
    :return: a function
    """
    def wrap(img):
        return img.normalizedDifference([green, swir2]).gt(threshold)
    return wrap


### DEPRECATED FUNCTIONS ###
# GENERIC APPLICATION OF MASKS

# LEDAPS
def ledaps(image):
    """ Function to use in Surface Reflectance Collections computed by
    LEDAPS

    Use:

    `masked = collection.map(cloud_mask.ledaps)`
    """
    cmask = image.select('QA')

    valid_data_mask = tools.image.computeBits(cmask, 1, 1, 'valid_data')
    cloud_mask = tools.image.computeBits(cmask, 2, 2, 'cloud')
    snow_mask = tools.image.computeBits(cmask, 4, 4, 'snow')

    good_pix = cloud_mask.eq(0).And(valid_data_mask.eq(0)).And(snow_mask.eq(0))
    result = image.updateMask(good_pix)

    return result


def landsatSR(options=('cloud', 'shadow', 'adjacent', 'snow'), name='sr_mask',
              addBands=False, updateMask=True):
    """ Function to use in Landsat Surface Reflectance Collections:
    LANDSAT/LT04/C01/T1_SR, LANDSAT/LT05/C01/T1_SR, LANDSAT/LE07/C01/T1_SR,
    LANDSAT/LC08/C01/T1_SR

    :param options: masks to apply. Options: 'cloud', 'shadow', 'adjacent',
        'snow'
    :type options: list
    :param name: name of the band that will hold the final mask. Default: 'toa_mask'
    :type name: str
    :param addBands: add all bands to the image. Default: False
    :type addBands: bool
    :param updateMask: update the mask of the Image. Default: True
    :type updateMask: bool
    :return: a function for applying the mask
    :rtype: function
    """
    sr = {'bits': ee.Dictionary({'cloud': 1, 'shadow': 2, 'adjacent': 3, 'snow': 4}),
          'band': 'sr_cloud_qa'}

    pix = {'bits': ee.Dictionary({'cloud': 5, 'shadow': 3, 'snow': 4}),
           'band': 'pixel_qa'}

    # Parameters
    options = ee.List(options)

    def wrap(image):
        bands = image.bandNames()
        contains_sr = bands.contains('sr_cloud_qa')
        good_pix = ee.Image(ee.Algorithms.If(contains_sr,
                   compute(image, sr['band'], sr['bits'], options, name_all=name),
                   compute(image, pix['band'], pix['bits'], options, name_all=name)))

        mask = good_pix.select([name]).Not()

        if addBands and updateMask:
            return image.updateMask(mask).addBands(good_pix)
        elif addBands:
            return image.addBands(good_pix)
        elif updateMask:
            return image.updateMask(mask)
        else:
            return image

    return wrap




# coding=utf-8
""" Module holding tools for ee.Date """
import ee
from datetime import datetime, timedelta

EE_EPOCH = datetime(1970, 1, 1, 0, 0, 0)


def toDatetime(date):
    """ convert a `ee.Date` into a `datetime` object """
    formatted = date.format('yyyy,MM,dd,HH,mm,ss').getInfo()
    args = formatted.split(',')
    intargs = [int(arg) for arg in args]
    return datetime(*intargs)


def millisToDatetime(millis):
    """ Converts milliseconds from 1970-01-01T00:00:00 to a
    datetime object """
    seconds = millis/1000
    dt = timedelta(seconds=seconds)
    return EE_EPOCH + dt


def daterangeList(start_date, end_date, interval=1, unit='month'):
    """ Divide a range that goes from start_date to end_date into many
        ee.DateRange, each one holding as many units as the interval.

    :param start_date: the start date. For the second DateRange and the
        following, it'll be one second after the end of the previous DateRange
    :param end_date: the end date
    :param interval: range of the DateRange in units
    :param unit: can be 'year', 'month' 'week', 'day', 'hour', 'minute',
        or 'second'
    :return: a list holding ee.DateRange
    :rtype: list
    """
    units = ['year', 'month', 'week', 'day', 'hour', 'minute', 'second']
    if unit not in units:
        raise ValueError('unit param must be one of {}'.format(units))

    def callback(interval, unit):
        def wrap(n, ini):
            ini = ee.List(ini)
            last_range = ee.DateRange(ini.get(-1))
            last_date = last_range.end()
            next_date = last_date.advance(1, 'second')
            next_interval = next_date.advance(interval, unit)
            return ini.add(ee.DateRange(next_date, next_interval))
        return wrap

    total_days = end_date.difference(start_date, unit)
    total_dateranges = total_days.divide(interval).toInt()
    dateranges_list = ee.List.sequence(1, total_dateranges)

    # first daterange
    first = ee.DateRange(start_date, start_date.advance(interval, unit))

    return ee.List(dateranges_list.iterate(callback(interval, unit),
                                           ee.List([first])))


def daterangeIntervals(start_date, end_date, interval=1, unit='month',
                       date_range=(1, 1), date_range_unit='day',
                       direction='backward'):
    """ Make Date Ranges in equal intervals

    :param interval: the amount of the interval in `unit`
    :param unit: the unit of the interval
    :param date_range: the amount of `date_range_unit` to go
        backwards (item 0) and forward (item 1)
    :param date_range_unit: the unit for date_range
    :param direction: the direction to advance
    """
    start_date = ee.Date(start_date)
    end_date = ee.Date(end_date)
    amplitude = end_date.difference(start_date, unit).round()
    intervals = amplitude.divide(interval).toInt()
    proxy = ee.List.sequence(0, intervals)

    if direction == 'forward':
        dates = proxy.map(
            lambda i: start_date.advance(
                ee.Number(i).multiply(interval), unit))
    else:
        dates = proxy.map(
            lambda i: end_date.advance(
                ee.Number(i).multiply(-interval), unit))

    def make_drange(date):
        date = ee.Date(date)
        return ee.DateRange(
            date.advance(-date_range[0], date_range_unit),
            date.advance(date_range[1], date_range_unit))

    return dates.map(lambda d: make_drange(d))


def dayRangeIntervals(start, end, interval=30, reverse=False, buffer='second'):
    """ Divide a date range into many DateRange objects. Return a list of DateRange

    :param start: the start date
    :param end: the end date
    :param interval: range of the DateRange in units
    :return: a list holding ee.DateRange
    :rtype: list
    """
    unit = 'day'
    start_date = ee.Date(start)
    end_date = ee.Date(end)
    begin_date = end_date if reverse else start_date

    diff = end_date.difference(start_date, unit)
    complete_intervals = diff.divide(interval).toInt()

    start_i_list = ee.List.sequence(0, complete_intervals)\
                          .map(lambda i: ee.Number(i).multiply(interval))
    end_i_list = ee.List.sequence(1, complete_intervals)\
                        .map(lambda i: ee.Number(i).multiply(interval)).add(diff)

    if reverse:
        start_i_list = start_i_list.map(lambda i: ee.Number(i).multiply(-1))
        end_i_list = end_i_list.map(lambda i: ee.Number(i).multiply(-1))

    start_end_i_list = start_i_list.zip(end_i_list)

    def wrap(z):
        z = ee.List(z)
        si = ee.Number(z.get(1 if reverse else 0))
        ei = ee.Number(z.get(0 if reverse else 1))
        start = begin_date.advance(si, unit)
        end = begin_date.advance(ei, unit)
        # apply buffer
        end = end.advance(1, buffer)
        return ee.DateRange(start, end)

    final = start_end_i_list.map(wrap)
    return final.reverse() if reverse else final


def unitSinceEpoch(date, unit='day'):
    """ Return the number of units since the epoch (1970-1-1)

    :param date: the date
    :type date: ee.Date
    :param unit: one of 'year', 'month' 'week', 'day', 'hour', 'minute',
        or 'second'
    :return: the corresponding units from the epoch
    :rtype: ee.Number
    """
    epoch = ee.Date(EE_EPOCH.isoformat())
    return date.difference(epoch, unit).toInt()


def getDateBand(img, unit='day', bandname='date', property_name=None):
    """ Get a date band from an image representing units since epoch

    :param img: the Image
    :param unit: one of 'year', 'month' 'week', 'day', 'hour', 'minute',
        or 'second'
    :param bandname: the name of the resulting band
    :return: a single band image with the date as the value for each pixel
        and also as an attribute
    :rtype: ee.Image
    """
    date = img.date()
    diff = unitSinceEpoch(date, unit)
    datei = ee.Image.constant(diff).rename(bandname)
    if not property_name:
        property_name = '{}_since_epoch'.format(unit)

    datei_attr = datei.set(property_name, diff).toInt()

    return datei_attr.copyProperties(img, ['system:footprint'])


def makeDateBand(image, format='YMMdd', bandname='date'):
    """ Make a date band using a formatter. Format pattern:

    C       century of era (>=0)         number        20
    Y       year of era (>=0)            year          1996

    x       weekyear                     year          1996
    w       week of weekyear             number        8
    ww      week of weekyear             number        08
    e       day of week                  number        2
    ee      day of week                  number        02

    y       year                         year          1996
    D       day of year                  number        5
    DD      day of year                  number        05
    DDD     day of year                  number        005
    M       month of year                month         7
    MM      month of year                month         07
    d       day of month                 number        5
    dd      day of month                 number        05

    K       hour of halfday (0~11)       number        0
    h       clockhour of halfday (1~12)  number        12

    H       hour of day (0~23)           number        0
    k       clockhour of day (1~24)      number        24
    m       minute of hour               number        30
    s       second of minute             number        55
    S       fraction of second           number        978

    Use the image first band for setting the resulting band projection
    """
    f = ee.String(format)
    # catch string formats for month
    pattern = f.replace('(MMM+)', 'MM')

    proj = image.select(0).projection()

    footprint = image.geometry()

    idate = image.date().format(pattern)
    idate_number = ee.Number.parse(idate)
    date_band = ee.Image.constant(idate_number).rename(bandname)
    date_band = date_band.toInt()  # force to be an Integer
    final = ee.Image(ee.Algorithms.If(footprint.isUnbounded(), date_band,
                                      date_band.clip(footprint)))

    return final.setDefaultProjection(proj)


def regularIntervals(start_date, end_date, interval=1, unit='month',
                     date_range=(1, 1), date_range_unit='day',
                     direction='backward'):
    """ Make date ranges at regular intervals

    :param start_date: if `direction` is forward the intervals will
        start from this date
    :type start_date: str or ee.Date
    :param end_date: if `direction` is backward the intervals will
        start from this date
    :type end_date: str or ee.Date
    :param interval: the interval
    :type interval: int
    :param unit: the unit for the interval
    :type unit: str
    :param date_range: the interval for the date range. It has to be a
        tuple or list in which the first item is for going backwards in
        `date_range_unit` and the second item for going forward
    :type date_range: tuple or list
    :param date_range_unit: the unit for each date range
    :type date_range_unit: str
    :param direction: it can be 'forward' or 'backward'
    :type direction: str
    :return: a list of date ranges
    :rtype: ee.List
    """
    start_date = ee.Date(start_date)
    end_date = ee.Date(end_date)
    amplitude = end_date.difference(start_date, unit).round()
    intervals = amplitude.divide(interval).toInt()
    proxy = ee.List.sequence(0, intervals)

    if direction == 'forward':
        dates = proxy.map(lambda i: start_date.advance(ee.Number(i).multiply(interval), unit))
    else:
        dates = proxy.map(lambda i: end_date.advance(ee.Number(i).multiply(-interval), unit))

    def make_drange(date):
        date = ee.Date(date)
        return ee.DateRange(date.advance(-date_range[0], date_range_unit),
                            date.advance(date_range[1], date_range_unit))

    return dates.map(lambda d: make_drange(d))


def dateSinceEpoch(date, unit='day'):
    """ Get the date for the specified date in unit

    :param date: the date in the specified unit
    :type date: int
    :param unit: one of 'year', 'month' 'week', 'day', 'hour', 'minute',
        or 'second'
    :return: the corresponding date
    :rtype: ee.Date
    """
    epoch = ee.Date(EE_EPOCH.isoformat())
    return epoch.advance(date, unit)


def fromDOY(doy, year):
    """ Creat a ee.Date given a Day of Year and a Year """
    def less10(doy):
        doy = doy.toInt()
        return ee.String('00').cat(ee.Number(doy).format())

    def less100(doy):
        doy = doy.toInt()
        return ee.String('0').cat(ee.Number(doy).format())

    doy = ee.Number(doy).add(1).toInt()
    year_str = ee.Number(year).format()
    doy_str = ee.Algorithms.If(doy.lt(10), less10(doy),
                               ee.String(ee.Algorithms.If(doy.lt(100), less100(doy), doy.format())))
    s = ee.String(doy_str).cat(year_str)

    return ee.Date.parse('DDDyyyy', s)


def isLeap(year):
    """ Determine wheater a year is leap or not. Returns 1 if leap, 0 if not """
    year = ee.Number(year)

    divisible4 = year.mod(4).gt(0)
    divisible100 = year.mod(100).gt(0)
    divisible400 = year.mod(400).gt(0)

    leap = ee.Algorithms.If(
        divisible4, 0,
        ee.Algorithms.If(
            divisible100, 1,
            ee.Algorithms.If(
                divisible400, 0, 1)))
    return ee.Number(leap)





# -*- coding: utf-8 -*-
""" Expression generator for Google Earth Engine """
from __future__ import print_function


class Expression(object):
    def __init__(self):
        pass

    @staticmethod
    def max(a, b):
        """ Generates the expression for max(a, b)

        :param a: one value. Can be a number or a variable
        :type a: str
        :param b: the other value
        :type b: str
        :return: max number
        :rtype: str
        """
        return "({a}>{b})?{a}:{b}".format(a= a, b= b)

    @staticmethod
    def min(a, b):
        """ Generates the expression for min(a, b)

        :param a: one value. Can be a number or a variable
        :type a: str
        :param b: the other value
        :type b: str
        :return: max number
        :rtype: str
        """
        return "({a}>{b})?{b}:{a}".format(a= a, b= b)


# coding=utf-8
""" Manager module for file managment """

import ee
import ee.data
import os
import json


def esave(eeobject, filename, path=None):
    """ Saves any EE object to a file with extension .gee

        The file has to be opened with `eopen`
    """
    obj = ee.serializer.encode(eeobject)

    path = path if path else os.getcwd()

    with open(os.path.join(path, filename+'.gee'), 'w') as js:
        json.dump(obj, js)


def eopen(file, path=None):
    """ Opens a files saved with `esave` method

    :return: the EE object """

    path = path if path else os.getcwd()

    try:
        with open(os.path.join(path, file), 'r') as gee:
            thefile = json.load(gee)
    except IOError:
        with open(os.path.join(path, file+'.gee'), 'r') as gee:
            thefile = json.load(gee)

    return ee.deserializer.decode(thefile)


# coding=utf-8
""" Module holding tools for ee.ImageCollections and ee.FeatueCollections """
import ee
import ee.data
from collections import OrderedDict


def extractList(dict, list):
    """ Extract values from a list of keys """
    empty = ee.List([])
    list = ee.List(list)
    dict = ee.Dictionary(dict)
    def iteration(el, first):
        f = ee.List(first)
        cond = dict.contains(el)
        return ee.Algorithms.If(cond, f.add(dict.get(el)), f)
    values = ee.List(list.iterate(iteration, empty))
    return values


def fromList(alist):
    """ Create a ee.Dictionary from a list of [[key, val], [key2, val2]...] """
    l = ee.List(alist)
    empty = ee.Dictionary({})
    def overList(ll, e):
        e = ee.Dictionary(e)
        ll = ee.List(ll)
        key = ll.get(0)
        val = ll.get(1)
        return e.set(key, val)
    return ee.Dictionary(l.iterate(overList, empty))


def sort(dictionary):
    """ Sort a dictionary. Can be a `dict` or a `ee.Dictionary`

    :param dictionary: the dictionary to sort
    :type dictionary: dict or ee.Dictionary
    :rtype: OrderedDict or ee.Dictionary
    """
    if isinstance(dictionary, dict):
        sorted = OrderedDict()
        keys = list(dictionary.keys())
        keys.sort()
        for key in keys:
            sorted[key] = dictionary[key]
        return sorted
    elif isinstance(dictionary, ee.Dictionary):
        keys = dictionary.keys()
        ordered = keys.sort()

        def iteration(key, first):
            new = ee.Dictionary(first)
            val = dictionary.get(key)
            return new.set(key, val)

        return ee.Dictionary(ordered.iterate(iteration, ee.Dictionary()))
    else:
        return dictionary


# coding=utf-8
""" User Interface Tools """
import ee
import threading
import pprint
from . import dispatcher, map

ASYNC = False


def eprint(*args, **kwargs):
    """ Print EE Objects. Similar to `print(object.getInfo())` but with
    some magic (lol)

    :param eeobject: object to print
    :type eeobject: ee.ComputedObject
    :param indent: indentation of the print output
    :type indent: int
    :param do_async: call getInfo() asynchronously
    :type do_async: bool
    """
    indent = kwargs.get('indent', 2)
    do_async = kwargs.get('do_async', ASYNC)
    pp = pprint.PrettyPrinter(indent=indent)

    info_return = [None]*len(args)

    def get_info(eeobject, index):
        """ Get Info """
        info_return[index] = dispatcher.dispatch(eeobject)

    for i, eeobject in enumerate(args):
        # DO THE SAME FOR EVERY OBJECT
        if do_async:
            thread = threading.Thread(target=get_info,
                                      args=(eeobject, i))
            thread.start()
        else:
            get_info(eeobject, i)

    for result in info_return:
        pp.pprint(result)


def getInfo(eeobject):
    """ Get eeobject information (getInfo) asynchronously. For not async just
    use `ee.data.getInfo` """

    class newDict(dict):
        def get(self):
            return self['info']
        def __call__(self):
            return self.get()

    result = newDict({'info':None})

    def get_info(eeobject, from_ee):
        if from_ee:
            info = eeobject.getInfo()
        else:
            info = eeobject

        result['info'] = info

    module = getattr(eeobject, '__module__', None)
    parent = module.split('.')[0] if module else None
    if parent == ee.__name__:
        thread = threading.Thread(target=get_info, args=(eeobject, True))
        thread.start()
    else:
        get_info(eeobject, False)

    return result



# coding=utf-8
""" Module holding tools for ee.FeatueCollections """
import ee
from . import collection as eecollection


def addId(collection, name='id', start=1):
    """ Add a unique numeric identifier, from parameter 'start' to
    collection.size() stored in a property called with parameter 'name'

    :param collection: the collection
    :type collection: ee.FeatureCollection
    :param name: the name of the resulting property
    :type name: str
    :param start: the number to start from
    :type start: int
    :return: the parsed collection with a new property
    :rtype: ee.FeatureCollection
    """
    start = ee.Number(start)
    collist = collection.toList(collection.size())
    first = ee.Feature(collist.get(0))
    rest = collist.slice(1)

    # Set first id
    first = ee.List([first.set(name, start)])

    # Set rest
    def over_col(feat, last):
        last = ee.List(last)
        last_feat = ee.Feature(last.get(-1))
        feat = ee.Feature(feat)
        last_id = ee.Number(last_feat.get('id'))
        return last.add(feat.set('id', last_id.add(1)))

    return ee.FeatureCollection(ee.List(rest.iterate(over_col, first)))


def enumerateProperty(col, name='enumeration'):
    """ Create a list of lists in which each element of the list is:
    [index, element]. For example, if you parse a FeatureCollection with 3
    Features you'll get: [[0, feat0], [1, feat1], [2, feat2]]

    :param collection: the collection
    :return: ee.FeatureCollection
    """
    enumerated = eecollection.enumerate(col)

    def over_list(l):
        l = ee.List(l)
        index = ee.Number(l.get(0))
        element = l.get(1)
        return ee.Feature(element).set(name, index)

    featlist = enumerated.map(over_list)
    return ee.FeatureCollection(featlist)


def enumerateSimple(collection, name='ENUM'):
    """ Simple enumeration of features inside a collection. Each feature stores
     its enumeration, so if the order of features changes over time, the
     numbers will not be in order """

    size = collection.size()
    collist = collection.toList(size)
    seq = ee.List.sequence(0, size.subtract(1))
    def wrap(n):
        n = ee.Number(n).toInt()
        feat = collist.get(n)
        return ee.Feature(feat).set(name, n)
    fc = ee.FeatureCollection(seq.map(wrap))

    return ee.FeatureCollection(fc.copyProperties(source=collection))


def listOptions(collection, propertyName):
    """ List all available values of `propertyName` in a feature collection """
    def wrap(feat, l):
        l = ee.List(l)
        return l.add(feat.get(propertyName))

    options = collection.iterate(wrap, ee.List([]))

    return ee.List(options).distinct()


def mergeGeometries(collection):
    """ Merge the geometries of many features. Return ee.Geometry """
    alist = collection.toList(collection.size())

    first = ee.Feature(alist.get(0))
    rest = alist.slice(1)

    def wrap(feat, ini):
        ini = ee.Geometry(ini)
        feat = ee.Feature(feat)
        geom = feat.geometry()
        union = geom.union(ini)
        return union.dissolve()

    return ee.Geometry(rest.iterate(wrap, first.geometry()))


# coding=utf-8
from __future__ import absolute_import
from ._version import __version__

__title__ = "geetools"
__summary__ = "A set of useful tools to use with Google Earth Engine Python" \
              "API"
__uri__ = "http://geetools.readthedocs.io"

__author__ = "Rodrigo E. Principe"
__email__ = "fitoprincipe82@gmail.com"

__license__ = "MIT"
__copyright__ = "2017 Rodrigo E. Principe"

try:
    from . import tools, bitreader, cloud_mask, expressions, decision_tree,\
                  filters, indices, batch, algorithms, composite,\
                  manager, utils, collection, oauth, visualization
    from .tools import array, date, dictionary, ee_list, featurecollection, \
                       geometry, image, imagecollection, number, string
    from .ui import eprint
    from .batch import Export, Import, Convert, Download
    from .oauth import Initialize
except ImportError:
    pass



# coding=utf-8
""" Tools for Earth Engine ee.List objects """
import ee
from . import dictionary, ee_list


def eq(string, to_compare):
    """ Compare two ee.String and return 1 if equal else 0 """
    string = ee.String(string)
    to_compare = ee.String(to_compare)
    return string.compareTo(to_compare).Not()


def format(string, replacement):
    """ Format a string using variables (as str.format). You can format numbers
    using the pattern: `{var:format}` where format string must be according to
    the argument in function ee.Number.format, which is based on
    https://docs.oracle.com/javase/7/docs/api/java/util/Formatter.html

    Example:

        string = ee.String('hello {nn:%.2f} {ll} {something} {else}')
        replacement = {'something': 'world', 'else': 'people',
                       'nn': 1.2555555654, 'pp': 'ignore'}

        formatted = geetools.tools.string.format(string, replacement)

        >
    """
    # casts
    replacement = ee.Dictionary(replacement)
    s = ee.String(string)
    # get keywords from string
    match = ee.String(string).match('{.*?}', 'g')
    match = match.map(lambda s: ee.String(s).slice(1,-1))

    # get keywords from replacement
    keys = replacement.keys()

    def addFormat(st):
        st = ee.String(st)
        splitted = ee.List(st.split(':'))
        length = splitted.size()
        cond = length.eq(1)

        def true():
            return splitted.add('0')
        def false():
            proxy = splitted.set(0, st)
            return splitted

        return ee.List(ee.Algorithms.If(cond, true(), false()))

    formats = dictionary.fromList(match.map(addFormat))
    selection = ee_list.intersection(keys, formats.keys())

    # filter replacement by the intersection keywords
    repl = replacement.select(selection)
    keys = repl.keys()

    def formatValues(k, val):
        f = ee.String(formats.get(k))
        def true(v):
            return ee.Number(val).format(f)
        def false(v):
            return ee.Algorithms.String(v)
        cond = f.compareTo('0') # returns 0 if equals
        return ee.String(ee.Algorithms.If(cond, true(val), false(val)))

    repl = repl.map(formatValues)
    values = repl.values()#.map(lambda v: ee.Algorithms.String(v))
    # get a list of lists [key, value]
    # example: [['foo', 'foovar'], ['bar', 'barvar']]
    z = keys.zip(values)

    def wrap(kv, ini):
        keyval = ee.List(kv)
        ini = ee.String(ini)

        key = ee.String(keyval.get(0))
        value = ee.String(keyval.get(1))

        pattern = ee.String('{').cat(key).cat(ee.String('.*?}'))

        return ini.replace(pattern, value, 'g')

    newstr = z.iterate(wrap, s)
    return ee.String(newstr)


def _zip(l1, l2):
    l1 = ee_list.toString(l1)
    l2 = ee_list.toString(l2)
    def wrap(el):
        el = ee.String(el)
        return l2.map(lambda e: el.cat(ee.String(e)))
    return l1.map(wrap).flatten()


def mix(strings):
    """ Mix a list of lists. For example:

    [[1,2,3], ['a', '1', 4]] -> ['1a', '11', '14', '2a', '21', '24', '3a', '31', '34']
    """
    strings = ee.List(strings)
    first = ee.List([ee.List(strings.get(0))])
    rest = ee.List(strings.slice(1))

    def wrap(eelist, aggregate):
        aggregate = ee.List(aggregate)
        last = ee.List(aggregate.get(-1))
        return ee.List([_zip(last, eelist)])

    return ee.List(rest.iterate(wrap, first)).flatten()




# coding=utf-8
""" Module holding tools for creating composites """
import ee
from uuid import uuid4
from . import tools, algorithms


def medoidScore(collection, bands=None, discard_zeros=False,
                bandname='sumdist', normalize=True):
    """ Compute a score to reflect 'how far' is from the medoid. Same params
     as medoid() """
    first_image = ee.Image(collection.first())
    if not bands:
        bands = first_image.bandNames()

    # Create a unique id property called 'enumeration'
    enumerated = tools.imagecollection.enumerateProperty(collection)
    collist = enumerated.toList(enumerated.size())

    def over_list(im):
        im = ee.Image(im)
        n = ee.Number(im.get('enumeration'))

        # Remove the current image from the collection
        filtered = tools.ee_list.removeIndex(collist, n)

        # Select bands for medoid
        to_process = im.select(bands)

        def over_collist(img):
            return ee.Image(img).select(bands)
        filtered = filtered.map(over_collist)

        # Compute the sum of the euclidean distance between the current image
        # and every image in the rest of the collection
        dist = algorithms.sumDistance(
            to_process, filtered,
            name=bandname,
            discard_zeros=discard_zeros)

        # Mask zero values
        if not normalize:
            # multiply by -1 to get the lowest value in the qualityMosaic
            dist = dist.multiply(-1)

        return im.addBands(dist)

    imlist = ee.List(collist.map(over_list))

    medcol = ee.ImageCollection.fromImages(imlist)

    # Normalize result to be between 0 and 1
    if normalize:
        min_sumdist = ee.Image(medcol.select(bandname).min())\
                        .rename('min_sumdist')
        max_sumdist = ee.Image(medcol.select(bandname).max()) \
                        .rename('max_sumdist')

        def to_normalize(img):
            sumdist = img.select(bandname)
            newband = ee.Image().expression(
                '1-((val-min)/(max-min))',
                {'val': sumdist,
                 'min': min_sumdist,
                 'max': max_sumdist}
            ).rename(bandname)
            return tools.image.replace(img, bandname, newband)

        medcol = medcol.map(to_normalize)

    return medcol


def medoid(collection, bands=None, discard_zeros=False):
    """ Medoid Composite. Adapted from https://www.mdpi.com/2072-4292/5/12/6481

    :param collection: the collection to composite
    :type collection: ee.ImageCollection
    :param bands: the bands to use for computation. The composite will include
        all bands
    :type bands: list
    :param discard_zeros: Masked and pixels with value zero will not be use
        for computation. Improves dark zones.
    :type discard_zeros: bool
    :return: the Medoid Composite
    :rtype: ee.Image
    """
    medcol = medoidScore(collection, bands, discard_zeros)
    comp = medcol.qualityMosaic('sumdist')
    final = tools.image.removeBands(comp, ['sumdist', 'mask'])
    return final


def closestDate(col, clip_to_first=False):
    """ Make a composite in which masked pixels are filled with the
    last available pixel. Make sure all image bands are casted

    :param clip_to_first: whether to clip with the 'first' image
        geometry
    """
    col = col.sort('system:time_start', False)
    first = ee.Image(col.first())

    # band names
    bandnames = first.bandNames()

    if clip_to_first:
        col = col.map(lambda img: img.clip(first.geometry()))

    tempname = 'a{}'.format(uuid4().hex)

    # add millis band (for compositing)
    col = col.map(lambda img: img.addBands(
        ee.Image.constant(img.date().millis()).rename(tempname).toInt()))

    col = col.sort('system:time_start')

    composite = col.qualityMosaic(tempname)

    return composite.select(bandnames).set('system:time_start',
                                           first.date().millis())


def compositeRegularIntervals(collection, interval=1, unit='month',
                              date_range=(1, 0), date_range_unit='month',
                              direction='backward', start=None, end=None,
                              composite_function=None,
                              composite_args=None,
                              composite_kwargs=None):
    """ Make a composite at regular intervals parsing a composite
    function. This function MUST return an ImageCollection and its first
    argument must be the input collection. The default function
    (if the argument is None) is `lambda col: col.median()`.
    """
    if composite_function is None:
        composite_function = lambda col: col.median()
    sorted_list = collection.sort('system:time_start').toList(
        collection.size())

    if start is None:
        start_date = ee.Image(sorted_list.get(0)).date()
    else:
        start_date = ee.Date(start)

    if end is None:
        end_date = ee.Image(sorted_list.get(-1)).date()
    else:
        end_date = ee.Date(end)

    date_ranges = tools.date.regularIntervals(
        start_date, end_date, interval, unit,
        date_range, date_range_unit, direction)

    def wrap(dr, l):
        l = ee.List(l)
        dr = ee.DateRange(dr)
        middle = ee.Number(dr.end().difference(dr.start(), 'day')).divide(2).floor()
        filtered = collection.filterDate(dr.start(), dr.end().advance(1, 'day'))
        dates = ee.List(filtered.aggregate_array('system:time_start')).map(lambda d: ee.Date(d).format())
        def true(filt, ll):
            if not composite_args and not composite_kwargs:
                comp = composite_function(filtered)
            elif composite_args and not composite_kwargs:
                comp = composite_function(filtered, *composite_args)
            elif composite_kwargs and not composite_args:
                comp = composite_function(filtered, **composite_kwargs)
            else:
                comp = composite_function(filtered, *composite_args, **composite_kwargs)

            comp = comp.set(
                'system:time_start', dr.start().advance(
                    middle, 'day').millis())
            comp = comp.set('dates', dates)\
                       .set('composite:time_start', dr.start().format())\
                       .set('composite:time_end', dr.end().format())
            return ll.add(comp)

        return ee.Algorithms.If(filtered.size(), true(filtered, l), l)

    return ee.ImageCollection.fromImages(
        ee.List(date_ranges.iterate(wrap, ee.List([]))))


def compositeByMonth(collection, composite_function=None, composite_args=None,
                     composite_kwargs=None):
    """ Make a composite at regular intervals parsing a composite
    function. This function MUST return an ImageCollection and its first
    argument must be the input collection. The default function
    (if the argument is None) is `lambda col: col.median()`.
    """
    if composite_function is None:
        composite_function = lambda col: col.median()

    years = ee.List(collection.aggregate_array('system:time_start'))\
              .map(lambda d: ee.Date(d).get('year')).distinct().sort()

    months = ee.List.sequence(1, 12)

    def wrapY(year):
        year = ee.Number(year)
        filteredY = collection.filter(ee.Filter.calendarRange(year, year, 'year'))

        def wrap(month, ilist):
            month = ee.Number(month)
            date = ee.Date.fromYMD(year, month, 1)
            filtered = filteredY.filter(ee.Filter.calendarRange(month, month, 'month'))
            def true(filt, ll):
                if not composite_args and not composite_kwargs:
                    comp = composite_function(filtered)
                elif composite_args and not composite_kwargs:
                    comp = composite_function(filtered, *composite_args)
                elif composite_kwargs and not composite_args:
                    comp = composite_function(filtered, **composite_kwargs)
                else:
                    comp = composite_function(filtered, *composite_args, **composite_kwargs)

                comp = comp.set(
                    'system:time_start', date.millis())
                return ee.List(ll).add(comp)
            return ee.Algorithms.If(filtered.size(), true(filtered, ilist), ilist)

        return ee.List(months.iterate(wrap, ee.List([])))

    images = ee.List(years.map(wrapY))
    return ee.ImageCollection.fromImages(images.flatten())


def max(collection, band):
    """ Make a max composite using the specified band """
    band = ee.String(band)
    first = collection.first()
    originalbands = first.bandNames()
    bands = originalbands.remove(band)
    bands = bands.insert(0, band)
    col = collection.map(lambda img: img.select(bands)) # change bands order
    comp = col.reduce(ee.Reducer.max(originalbands.size()))
    return comp.rename(bands).select(originalbands)


# coding=utf-8
""" Some util functions """

import pandas as pd
from copy import deepcopy
import ee
from .tools import string


def getReducerName(reducer):
    """
    Get the name of the parsed reducer.

    WARNING: This function makes a request to EE Servers (getInfo). Do not use
    in server side functions (example: inside a mapping function)
    """
    reducer_type = reducer.getInfo()['type']

    relations = dict(
        mean=['Reducer.mean', 'Reducer.intervalMean'],
        median=['Reducer.median'],
        mode=['Reducer.mode'],
        first=['Reducer.first', 'Reducer.firstNonNull'],
        last=['Reducer.last', 'Reducer.lastNonNull'],
        stdDev=['Reducer.stdDev', 'Reducer.sampleStdDev'],
        all=['Reducer.allNoneZero'],
        any=['Reducer.anyNoneZero'],
        count=['Reducer.count', 'Reducer.countDistinct'],
        max=['Reducer.max'],
        min=['Reducer.min'],
        product=['Reducer.product'],
        variance=['Reducer.sampleVariance', 'Reducer.variance'],
        skew=['Reducer.skew'],
        sum=['Reducer.sum']
    )

    for name, options in relations.items():
        if reducer_type in options:
            return name


def reduceRegionsPandas(data, index='system:index', add_coordinates=False,
                        duplicate_index=False):
    """ Transform data coming from Image.reduceRegions to a pandas dataframe

    :param data: data coming from Image.reduceRegions
    :type data: ee.Dictionary or dict
    :param index: the index of the dataframe
    :param add_coordinates: if True adds the coordinates to the dataframe
    :param duplicate_index: if True adds the index data to the dataframe too
    :return: a pandas dataframe
    :rtype: pd.DataFrame
    """
    if not isinstance(data, dict):
        if add_coordinates:
            def addCentroid(feat):
                feat = ee.Feature(feat)
                centroid = feat.centroid().geometry()
                coords = ee.List(centroid.coordinates())
                return feat.set('longitude', ee.Number(coords.get(0)),
                                'latitude', ee.Number(coords.get(1)))
            data = data.map(addCentroid)

        data = data.getInfo()

    features = data['features']

    d, indexes = [], []
    for feature in features:
        nf = deepcopy(feature)
        props = nf['properties']

        if not duplicate_index:
            props.pop(index) if index in props else props

        d.append(props)
        if index == 'system:index':
            indexes.append(feature['id'])
        else:
            indexes.append(feature['properties'][index])

    return pd.DataFrame(d, indexes)


def castImage(value):
    """ Cast a value into an ee.Image if it is not already """
    if isinstance(value, ee.Image) or value is None:
        return value
    else:
        return ee.Image.constant(value)


def makeName(img, pattern, date_pattern=None, extra=None):
    """ Make a name with the given pattern. The pattern must contain the
    propeties to replace between curly braces. There are 2 special words:

    * 'system_date': replace with the date of the image formatted with
      `date_pattern`, which defaults to 'yyyyMMdd'
    * 'id' or 'ID': the image id. If None, it'll be replaced with 'id'

    Pattern example (supposing each image has a property called `city`):
    'image from {city} on {system_date}'

    You can add extra parameters using keyword `extra`
    """
    img = ee.Image(img)
    props = img.toDictionary()
    props = ee.Dictionary(ee.Algorithms.If(
        img.id(),
        props.set('id', img.id()).set('ID', img.id()),
        props))
    props = ee.Dictionary(ee.Algorithms.If(
        img.propertyNames().contains('system:time_start'),
        props.set('system_date', img.date().format(date_pattern)),
        props))
    if extra:
        extra = ee.Dictionary(extra)
        props = props.combine(extra)
    name = string.format(pattern, props)

    return name


def maskIslands(mask, limit, pixels_limit=1000):
    """ returns a new mask where connected pixels with less than the 'limit'
    of area are turned to 0 """
    area = ee.Image.pixelArea().rename('area')

    conn = mask.connectedPixelCount(pixels_limit).rename('connected')
    finalarea = area.multiply(conn)

    # get holes and islands
    island = mask.eq(1).And(finalarea.lte(limit))

    # get rid island
    no_island = mask.where(island, 0)

    return no_island


def dict2namedtuple(thedict, name='NamedDict'):
    """ Create a namedtuple from a dict object. It handles nested dicts. If
    you want to scape this behaviour the dict must be placed into a list as its
    unique element """
    from collections import namedtuple

    thenametuple = namedtuple(name, [])

    for key, val in thedict.items():
        if not isinstance(key, str):
            msg = 'dict keys must be strings not {}'
            raise ValueError(msg.format(key.__class__))

        if not isinstance(val, dict):
            # workaround to include a dict as an attribute
            if isinstance(val, list):
                if isinstance(val[0], dict):
                    val = val[0]

            setattr(thenametuple, key, val)
        else:
            newname = dict2namedtuple(val, key)
            setattr(thenametuple, key, newname)

    return thenametuple


def formatVisParams(visParams):
    """ format visualization parameters to match EE requirements at
    ee.data.getMapId """
    formatted = dict()
    for param, value in visParams.items():
        if isinstance(value, list):
            value = [str(v) for v in value]
        if param in ['bands', 'palette']:
            formatted[param] = ','.join(value) if len(value) == 3 else str(value[0])
        if param in ['min', 'max', 'gain', 'bias', 'gamma']:
            formatted[param] = str(value) if isinstance(value, (int, str)) else ','.join(value)
    return formatted




# coding=utf-8
""" Basic Map tools """

import ee


def formatVisParams(visParams):
    """ format visualization params to match getMapId requirement """
    copy = {key: val for key, val in visParams.items()}

    def list2str(params):
        n = len(params)
        if n == 3:
            newbands = '{},{},{}'.format(params[0], params[1], params[2])
        else:
            newbands = '{}'.format(params[0])
        return newbands

    if 'bands' in copy:
        copy['bands'] = list2str(copy['bands'])

    if 'palette' in copy:
        if isinstance(copy['palette'], list):
            copy['palette'] = ','.join(copy['palette'])
        else:
            copy['palette'] = [str(copy['palette'])]

    for k in ['min', 'max', 'bias', 'gain', 'gamma']:
        if k in copy:
            if isinstance(copy[k], list):
                copy[k] = list2str(copy[k])
            else:
                value = float(copy[k])
                copy[k] = [value, value, value]

    return copy




# coding=utf-8
import ee
import os

GEOMETRY_TYPES = {
    'LineString': ee.geometry.Geometry.LineString,
    'LineRing': ee.geometry.Geometry.LinearRing,
    'MultiLineString': ee.geometry.Geometry.MultiLineString,
    'MultiPolygon': ee.geometry.Geometry.MultiPolygon,
    'MultiPoint': ee.geometry.Geometry.MultiPoint,
    'Point': ee.geometry.Geometry.Point,
    'Polygon': ee.geometry.Geometry.Polygon,
    'Rectangle': ee.geometry.Geometry.Rectangle,
    'GeometryCollection': ee.geometry.Geometry,
}


def getProjection(filename, path=None):
    """ Get EPSG from a shapefile using pycrs

    :param filename: an ESRI shapefile (.shp)
    :type filename: str
    """
    try:
        import pycrs
    except ModuleNotFoundError as e:
        print('Install pycrs mode by `pip install pycrs`')
        raise e
    except Exception as e:
        raise e
    import requests
    import os

    if not path:
        path = os.getcwd()

    BASEURL = 'http://prj2epsg.org/search.json'
    fname = filename.split('.')[0]
    prjname = '{}.prj'.format(fname)
    fpath = os.path.join(path, prjname)
    if not os.path.exists(prjname):
        raise ValueError('{} does not exist'.format(fpath))
    crs = pycrs.load.from_file(fpath)
    wkt = crs.to_ogc_wkt()
    params = dict(mode='wkt', terms=wkt)
    response = requests.get(BASEURL, params)
    rjson = response.json()
    return rjson['codes'][0]['code']


def kmlToGeoJsonDict(kmlfile=None, data=None, encoding=None):
    """ Convert a KML file to a GeoJSON dict """
    import xml.dom.minidom as md
    from fastkml import kml
    import kml2geojson

    k = kml.KML()

    with open(kmlfile) as thefile:
        kmlf = thefile.read()

    # Handle encoding
    if not encoding:
        try:
            import re
            match = re.search('encoding=".+"', kmlf).group()
            encoding = match.split('=')[1][1:-1]
        except:
            encoding = 'utf-8'

    kmlf = kmlf.encode(encoding)
    k.from_string(kmlf)
    kmlStr = k.to_string()

    # force encoding
    kmlStr = kmlStr.encode(encoding, errors="ignore").decode()
    root = md.parseString(kmlStr)
    layers = kml2geojson.build_feature_collection(root)
    return layers


def isPoint(pointlist):
    """ Verify is a list is a list of points """
    if len(pointlist) in [2, 3]:
        if isinstance(pointlist[0], (int, float))\
                and isinstance(pointlist[1], (int, float)):
            return True
        else:
            return False
    else:
        return False

def hasZ(pointlist):
    """ determine if points inside coordinates have Z values """
    points = pointlist[0]
    first = points[0]
    if len(first) == 3:
        return True
    else:
        return False


def removeZ(coords):
    """ Remove Z values from coordinates """
    newcoords = coords.copy()
    for p in newcoords[0]:
        p.pop(2)
    return newcoords


def recrusiveDeleteAsset(assetId):
    info = ee.data.getInfo(assetId)
    if info:
        ty = info['type']
        if ty in ['Image', 'FeatureCollection']:
            # setting content to 0 will delete the assetId
            content = 0
        elif ty in ['Folder', 'ImageCollection']:
            try:
                content = ee.data.getList({'id':assetId})
            except Exception as e:
                print(str(e))
                return
        else:
            print("Can't handle {} type yet".format(ty))

        if content == 0:
            # delete empty colletion and/or folder
            ee.data.deleteAsset(assetId)
        else:
            for asset in content:
                path = asset['id']
                ty = asset['type']
                if ty == 'Image':
                    # print('deleting {}'.format(path))
                    ee.data.deleteAsset(path)
                else:
                    recrusiveDeleteAsset(path)
            # delete empty collection and/or folder
            ee.data.deleteAsset(assetId)
    else:
        print('{} does not exists or there is another problem'.format(assetId))


def convertDataType(newtype):
    """ Convert an image to the specified data type

    :param newtype: the data type. One of 'float', 'int', 'byte', 'double',
        'Uint8','int8','Uint16', 'int16', 'Uint32','int32'
    :type newtype: str
    :return: a function to map over a collection
    :rtype: function
    """
    newtype = newtype.lower()
    def wrap(image):
        TYPES = {'float': image.toFloat,
                 'int': image.toInt,
                 'byte': image.toByte,
                 'double': image.toDouble,
                 'uint8': image.toUint8,
                 'int8': image.toInt8,
                 'uint16': image.toUint16,
                 'int16': image.toInt16,
                 'uint32': image.toUint32,
                 'int32': image.toInt32}
        return TYPES[newtype]()
    return wrap


def create_asset(asset_id, asset_type, mk_parents=True):
    """ Create an Asset """
    types = {
        'ImageCollection': 'IMAGE_COLLECTION',
        'Folder': 'FOLDER'
    }

    already = ee.data.getInfo(asset_id)
    if already:
        ty = already['type']
        if ty != types[asset_type]:
            raise ValueError("{} is a {}. Can't create asset".format(asset_id, ty))
        else:
            return None

    if mk_parents:
        parts = asset_id.split('/')
        root = "/".join(parts[:2])
        root += "/"
        for part in parts[2:-1]:
            root += part
            if ee.data.getInfo(root) is None:
                ee.data.createAsset({'type': 'Folder'}, root)
            root += '/'
    return ee.data.createAsset({'type': asset_type}, asset_id)


def createAssets(asset_ids, asset_type, mk_parents):
    """Creates the specified assets if they do not exist.
    This is a fork of the original function in 'ee.data' module with the
    difference that

    - If the asset already exists but the type is different that the one we
      want, raise an error
    - Starts the creation of folders since 'user/username/'

    Will be here until I can pull requests to the original repo

    :param asset_ids: list of paths
    :type asset_ids: list
    :param asset_type: the type of the assets. Options: "ImageCollection" or
        "Folder"
    :type asset_type: str
    :param mk_parents: make the parents?
    :type mk_parents: bool
    :return: A description of the saved asset, including a generated ID

    """
    for asset_id in asset_ids:
        already = ee.data.getInfo(asset_id)
        if already:
            ty = already['type']
            if ty != asset_type:
                raise ValueError("{} is a {}. Can't create asset".format(asset_id, ty))
            print('Asset %s already exists' % asset_id)
            continue
        if mk_parents:
            parts = asset_id.split('/')
            root = "/".join(parts[:2])
            root += "/"
            for part in parts[2:-1]:
                root += part
                if ee.data.getInfo(root) is None:
                    ee.data.createAsset({'type': 'Folder'}, root)
                root += '/'
        return ee.data.createAsset({'type': asset_type}, asset_id)


def downloadFile(url, name, extension, path=None):
    """ Download a file from a given url

    :param url: full url
    :type url: str
    :param name: name for the file (can contain a path)
    :type name: str
    :param extension: extension for the file
    :type extension: str
    :return: the created file (closed)
    :rtype: file
    """
    import requests
    response = requests.get(url, stream=True)
    code = response.status_code

    if path is None:
        path = os.getcwd()

    pathname = os.path.join(path, name)

    while code != 200:
        if code == 400:
            return None
        response = requests.get(url, stream=True)
        code = response.status_code
        size = response.headers.get('content-length',0)
        if size: print('size:', size)

    with open('{}.{}'.format(pathname, extension), "wb") as handle:
        for data in response.iter_content():
            handle.write(data)

    return handle


def matchDescription(name, custom=None):
    """ Format a name to be accepted as a desciption.
    The rule is:

    The description must contain only the following characters: a..z, A..Z,
    0..9, ".", ",", ":", ";", "_" or "-". The description must be at most 100
    characters long.
    """
    letters = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
               'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']
    numbers = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    upper = [s.capitalize() for s in letters]
    chars = ['.', ',', ':', ';', '_', '-']
    allchars  = letters+upper+chars+numbers

    replacements = [
        [' ', [' ']],
        ['-', ['/']],
        ['.', ['?', '!', '¿', '*']],
        [':', ['(', ')','[', ']', '{', '}']],
        ['a', ['á', 'ä', 'à', 'æ']],
        ['e', ['é', 'ë', 'è']],
        ['i', ['í', 'ï', 'ì']],
        ['o', ['ó', 'ö', 'ò', 'ø']],
        ['u', ['ú', 'ü', 'ù']],
        ['c', ['¢', 'ç']],
        ['n', ['ñ']]
    ]

    replacementupper = []
    for r in replacements:
        row = []
        row.append(r[0].capitalize())
        row2 = []
        for alt in r[1]:
            row2.append(alt.capitalize())
        row.append(row2)
        replacementupper.append(row)

    replacements_dict = dict()
    for replacement in replacements+replacementupper:
        letter = replacement[0]
        repl = replacement[1]
        for char in repl:
            replacements_dict[char] = letter

    # user custom mapping
    if custom:
        replacements_dict.update(custom)

    description = ''
    for letter in name:
        if letter not in allchars:
            if letter in replacements_dict:
                description += replacements_dict[letter]
            else:
                description += ''
        else:
            description += letter

    return description[0:100]


# coding=utf-8
""" Module holding tools for ee.Number """
import ee
import ee.data


def trimDecimals(number, places=2):
    """ Decrease the number of decimals in a ee.Number

    :param places: number of decimal places to leave
    :return: a function to map over a list
    """
    factor = ee.Number(10).pow(ee.Number(places).toInt())

    return number.multiply(factor).toInt().divide(factor)



# coding=utf-8

import ee
ee.Initialize()
from geetools import indices
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

image = ee.Image(TEST_CLOUD_IMAGES['S2'])
p = image.geometry().centroid()
col = ee.ImageCollection('COPERNICUS/S2').filterBounds(p)


def test_ndvi():
    collection = col.map(lambda img: img.addBands(indices.ndvi(img, 'B8','B4')))
    image = ee.Image(collection.first())
    values = getValue(image, p, 10, side='client')
    NIR = float(values['B8'])
    RED = float(values['B4'])
    index = (NIR-RED)/(NIR+RED)
    index_from_i = getValue(image, p, 10, 'client')['ndvi']

    assert index == index_from_i


def test_evi():
    collection = col.map(lambda img: img.addBands(indices.evi(img, 'B8', 'B4', 'B2')))
    image = ee.Image(collection.first())
    values = getValue(image, p, 10, side='client')
    NIR = float(values['B8'])
    RED = float(values['B4'])
    BLUE = float(values['B2'])

    index = (2.5)*((NIR-RED)/(NIR+(6*RED)-((7.5)*BLUE)+1))
    index_from_i = getValue(image, p, 10, 'client')['evi']

    assert index == index_from_i


def test_nbr():
    collection = col.map(lambda img: img.addBands(indices.nbr(img, 'B8', 'B12')))
    image = ee.Image(collection.first())
    values = getValue(image, p, 10, side='client')
    NIR = float(values['B8'])
    SWIR2 = float(values['B12'])

    index = (NIR-SWIR2)/(NIR+SWIR2)
    index_from_i = getValue(image, p, 10, 'client')['nbr']

    assert index == index_from_i


def test_nbr2():
    collection = col.map(lambda img: img.addBands(indices.nbr2(img, 'B11', 'B12')))
    image = ee.Image(collection.first())
    values = getValue(image, p, 10, side='client')
    SWIR1 = float(values['B11'])
    SWIR2 = float(values['B12'])

    index = (SWIR1-SWIR2)/(SWIR1+SWIR2)
    index_from_i = getValue(image, p, 10, 'client')['nbr2']

    assert index == index_from_i


# -*- coding: utf-8 -*-

import ee
ee.Initialize()
from geetools import tools


def test_daterange_list():
    start_date = ee.Date('2010-01-01')
    end_date = ee.Date('2015-01-01')
    unit = 'year'
    interval = 1

    expected = ee.List([
        ee.DateRange('2010-01-01', '2011-01-01'),
        ee.DateRange('2011-01-01', '2012-01-01'),
        ee.DateRange('2012-01-01', '2013-01-01'),
        ee.DateRange('2013-01-01', '2014-01-01'),
        ee.DateRange('2014-01-01', '2015-01-01')
    ])

    daterange_list = tools.date.daterangeList(start_date, end_date,
                                              interval, unit)

    assert expected == daterange_list



# coding=utf-8

from geetools.tools import ee_list
import ee
ee.Initialize()

list1 = ee.List([1, 2, 3, 4, 5])
list2 = ee.List([4, 5, 6, 7])


# helper
def assert_equal(obj, compare):
    assert obj.getInfo() == compare


def test_list_intersection():
    intersection = ee_list.intersection(list1, list2)
    assert_equal(intersection, [4, 5])


def test_list_difference():
    diff = ee_list.difference(list1, list2)
    assert_equal(diff, [1, 2, 3, 6, 7])


def test_remove_duplicates():
    duplicated = ee.List([1, 2, 2, 1, 3, 5])
    unique = ee_list.removeDuplicates(duplicated)
    assert_equal(unique, [1, 2, 3, 5])


def test_replace_many():
    replaced = ee_list.replaceDict(list1, {1: 'test', 4:8})
    assert_equal(replaced, ['test', 2, 3, 8, 5])

    # new test to match image's test
    newlist = ee.List(["B1", "B2", "B3", "B4", "B5", "B6", "B7",
                        "cfmask", "cfmask_conf"])
    replaced = ee_list.replaceDict(newlist, {'B1': 'BLUE', 'B2': 'GREEN'})
    assert_equal(replaced, ["BLUE", "GREEN", "B3", "B4", "B5", "B6", "B7",
                            "cfmask", "cfmask_conf"])


def test_get_from_dict():
    test_list = ee.List(['a', 'c'])
    test_dict = ee.Dictionary({'a':1, 'b':5, 'c':8})

    values = ee_list.getFromDict(test_list, test_dict)

    # assert
    assert_equal(values, [1, 8])




# coding=utf-8
""" Module holding tools for ee.ImageCollections """
import ee
import ee.data
import pandas as pd
import math
from . import date
from . import image as image_module
from . import collection as eecollection
from ..utils import castImage
from .. import composite


def add(collection, image):
    """ Add an Image to the Collection

    **SERVER SIDE**

    """
    # TODO: handle a list of images
    collist = collection.toList(collection.size())
    append = collist.add(image)
    return ee.ImageCollection.fromImages(append)


def containsAllBands(collection, bands):
    """ Filter a collection with images cotaining all bands specified in
    parameter `bands` """
    bands = ee.List(bands)
    # add bands as metadata
    collection = collection.map(
        lambda i: ee.Image(i).set('_BANDS_', ee.Image(i).bandNames()))

    band0 = ee.String(bands.get(0))
    rest = ee.List(bands.slice(1))
    filt0 = ee.Filter.listContains(leftField='_BANDS_', rightValue=band0)

    # Get filter
    def wrap(band, filt):
        band = ee.String(band)
        filt = ee.Filter(filt)
        newfilt = ee.Filter.listContains(leftField='_BANDS_', rightValue=band)
        return ee.Filter.And(filt, newfilt)

    filt = ee.Filter(rest.iterate(wrap, filt0))
    return collection.filter(filt)


def containsAnyBand(collection, bands):
    """ Filter a collection with images cotaining any of the bands specified in
    parameter `bands` """
    bands = ee.List(bands)
    # add bands as metadata
    collection = collection.map(
        lambda i: ee.Image(i).set('_BANDS_', ee.Image(i).bandNames()))

    band0 = ee.String(bands.get(0))
    rest = ee.List(bands.slice(1))
    filt0 = ee.Filter.listContains(leftField='_BANDS_', rightValue=band0)

    # Get filter
    def wrap(band, filt):
        band = ee.String(band)
        filt = ee.Filter(filt)
        newfilt = ee.Filter.listContains(leftField='_BANDS_', rightValue=band)
        return ee.Filter.Or(filt, newfilt)

    filt = ee.Filter(rest.iterate(wrap, filt0))
    return collection.filter(filt)


def getId(collection):
    """ Get the ImageCollection id.

    **CLIENT SIDE**

    :type collection: ee.ImageCollection
    :return: the collection's id
    :rtype: str
    """
    return collection.limit(0).getInfo()['id']


def getImage(collection, index):
    """ Get an Image using its collection index """
    collist = collection.toList(collection.size())
    return ee.Image(collist.get(index))


def wrapper(f, *arg, **kwargs):
    """ Wrap a function and its arguments into a mapping function for
    ImageCollections. The first parameter of the functions must be an Image,
    and it must return an Image.

    :param f: the function to be wrapped
    :type f: function
    :return: a function to use in ee.ImageCollection.map
    :rtype: function
    """
    def wrap(img):
        return f(img, *arg, **kwargs)
    return wrap


def enumerateProperty(collection, name='enumeration'):
    """

    :param collection:
    :param name:
    :return:
    """
    enumerated = eecollection.enumerate(collection)

    def over_list(l):
        l = ee.List(l)
        index = ee.Number(l.get(0))
        element = l.get(1)
        return ee.Image(element).set(name, index)

    imlist = enumerated.map(over_list)
    return ee.ImageCollection(imlist)


def enumerateSimple(collection, name='ENUM'):
    """ Simple enumeration of features inside a collection. Each feature stores
     its enumeration, so if the order of features changes over time, the numbers
     will not be in order """
    size = collection.size()
    collist = collection.toList(size)
    seq = ee.List.sequence(0, size.subtract(1))
    def wrap(n):
        n = ee.Number(n).toInt()
        feat = collist.get(n)
        return ee.Image(feat).set(name, n)
    fc = ee.ImageCollection.fromImages(seq.map(wrap))

    return ee.ImageCollection(fc.copyProperties(source=collection))


def fillWithLast(collection, unmask_first=False):
    """ Fill masked values of each image pixel with the last available
    value

    :param collection: the collection that holds the images that will be filled
    :type collection: ee.ImageCollection
    :param unmask_first: unmask first image?
    :type unmask_first: bool
    :rtype: ee.ImageCollection
    """
    collector = ee.List([])
    def overcol(i, collect):
        collect = ee.List(collect)
        def true():
            last = ee.Image(collect.get(-1))
            mask = i.mask().Not()
            return collect.add(ee.Image(i.unmask().where(mask, last)))
        def false():
            return collect.add(i.unmask() if unmask_first else i)
        return ee.List(ee.Algorithms.If(collect.size(), true(), false()))
    return ee.ImageCollection.fromImages(collection.iterate(overcol, collector))


def mergeGeometries(collection):
    """ Merge the geometries of many images. Return ee.Geometry """
    imlist = collection.toList(collection.size())

    first = ee.Image(imlist.get(0))
    rest = imlist.slice(1)

    def wrap(img, ini):
        ini = ee.Geometry(ini)
        img = ee.Image(img)
        geom = img.geometry()
        union = geom.union(ini)
        return union.dissolve()

    return ee.Geometry(rest.iterate(wrap, first.geometry()))


def mosaicSameDay(collection, qualityBand=None):
    """ Return a collection where images from the same day are mosaicked

    :param qualityBand: the band that holds the quality score for mosaiking.
        If None it will use the simplier mosaic() function
    :type qualityBand: str
    :return: a new image collection with 1 image per day. The only property
        kept is `system:time_start`
    :rtype: ee.ImageCollection
    """
    all_dates = collection.aggregate_array('system:time_start')
    def overdates(d, l):
        l = ee.List(l)
        date = ee.Date(d)
        day = date.get('day')
        month = date.get('month')
        year = date.get('year')
        clean_date = ee.Date.fromYMD(year, month, day)
        condition = l.contains(clean_date)
        return ee.Algorithms.If(condition, l, l.add(clean_date))

    date_list = ee.List(all_dates.iterate(overdates, ee.List([])))
    first_img = ee.Image(collection.first())
    bands = first_img.bandNames()

    def make_col(date):
        date = ee.Date(date)
        filtered = collection.filterDate(date, date.advance(1, 'day'))

        if qualityBand:
            mosaic = filtered.qualityMosaic(qualityBand)
        else:
            mosaic = filtered.mosaic()

        mosaic = mosaic.set('system:time_start', date.millis(),
                            'system:footprint', mergeGeometries(filtered))

        # mosaic = mosaic.rename(bands)
        mosaic = mosaic.select(bands)
        def reproject(bname, mos):
            mos = ee.Image(mos)
            mos_bnames = mos.bandNames()
            bname = ee.String(bname)
            proj = first_img.select(bname).projection()

            newmos = ee.Image(ee.Algorithms.If(
                mos_bnames.contains(bname),
                image_module.replace(mos, bname, mos.select(bname).setDefaultProjection(proj)),
                mos))

            return newmos

        mosaic = ee.Image(bands.iterate(reproject, mosaic))
        return mosaic

    new_col = ee.ImageCollection.fromImages(date_list.map(make_col))
    return new_col


def reduceEqualInterval(collection, interval=30, unit='day', reducer=None,
                        start_date=None, end_date=None):
    """ Reduce an ImageCollection into a new one that has one image per
        reduced interval, for example, one image per month.

    :param collection: the collection
    :type collection: ee.ImageCollection
    :param interval: the interval to reduce
    :type interval: int
    :param unit: unit of the interval. Can be 'day', 'month', 'year'
    :param reducer: the reducer to apply where images overlap. If None, uses
        a median reducer
    :type reducer: ee.Reducer
    :param start_date: fix the start date. If None, uses the date of the first
        image in the collection
    :type start_date: ee.Date
    :param end_date: fix the end date. If None, uses the date of the last image
        in the collection
    :type end_date: ee.Date
    :return:
    """
    interval = int(interval)  # force to int
    first = ee.Image(collection.sort('system:time_start').first())
    bands = first.bandNames()

    if not start_date:
        start_date = first.date()
    if not end_date:
        last = ee.Image(collection.sort('system:time_start', False).first())
        end_date = last.date()
    if not reducer:
        reducer = ee.Reducer.median()

    def apply_reducer(red, col):
        return ee.Image(col.reduce(red))

    ranges = date.daterangeList(start_date, end_date, interval, unit)

    def over_ranges(drange, ini):
        ini = ee.List(ini)
        drange = ee.DateRange(drange)
        start = drange.start()
        end = drange.end()
        filtered = collection.filterDate(start, end)
        condition = ee.Number(filtered.size()).gt(0)
        def true():
            image = apply_reducer(reducer, filtered)\
                    .set('system:time_start', end.millis())\
                    .set('reduced_from', start.format())\
                    .set('reduced_to', end.format())
            # rename to original names
            image = image.select(image.bandNames(), bands)
            result = ini.add(image)
            return result
        return ee.List(ee.Algorithms.If(condition, true(), ini))

    imlist = ee.List(ranges.iterate(over_ranges, ee.List([])))

    return ee.ImageCollection.fromImages(imlist)


def makeEqualInterval(collection, interval=1, unit='month'):
    """ Make a list of image collections filtered by the given interval,
    for example, one month. Starts from the end of the parsed collection

    :param collection: the collection
    :type collection: ee.ImageCollection
    :param interval: the interval
    :type interval: int
    :param unit: unit of the interval. Can be 'day', 'month', 'year'
    :rtype: ee.List
    """
    interval = int(interval)  # force to int
    collist = collection.sort('system:time_start').toList(collection.size())
    start_date = ee.Image(collist.get(0)).date()
    end_date = ee.Image(collist.get(-1)).date()

    ranges = date.daterangeList(start_date, end_date, interval, unit)

    def over_ranges(drange, ini):
        ini = ee.List(ini)
        drange = ee.DateRange(drange)
        start = drange.start()
        end = drange.end()
        filtered = collection.filterDate(start, end)
        condition = ee.Number(filtered.size()).gt(0)
        return ee.List(ee.Algorithms.If(condition, ini.add(filtered), ini))

    imlist = ee.List(ranges.iterate(over_ranges, ee.List([])))

    return imlist


def makeDayIntervals(collection, interval=30, reverse=False, buffer='second'):
    """ Make day intervals """
    interval = int(interval)
    collection = collection.sort('system:time_start', True)
    start = collection.first().date()
    end = collection.sort('system:time_start', False).first().date()
    ranges = date.dayRangeIntervals(start, end, interval, reverse, buffer)

    def over_ranges(drange, ini):
        ini = ee.List(ini)
        drange = ee.DateRange(drange)
        start = drange.start()
        end = drange.end()
        filtered = collection.filterDate(start, end)
        condition = ee.Number(filtered.size()).gt(0)
        return ee.List(ee.Algorithms.If(condition, ini.add(filtered), ini))

    imlist = ee.List(ranges.iterate(over_ranges, ee.List([])))

    return imlist


def reduceDayIntervals(collection, reducer, interval=30, reverse=False,
                       buffer='second'):
    """ Reduce Day Intervals

    :param reducer: a function that takes as only argument a collection
        and returns an image
    :type reducer: function
    :return: an image collection
    :rtype: ee.ImageCollection
    """
    intervals = makeDayIntervals(collection, interval, reverse, buffer)
    reduced = intervals.map(reducer)
    return ee.ImageCollection.fromImages(reduced)


def getValues(collection, geometry, scale=None, reducer=None,
              id='system:index', properties=None, side='server',
              maxPixels=1e9):
    """ Return all values of all bands of an image collection in the
        specified geometry

    :param geometry: Point from where to get the info
    :type geometry: ee.Geometry
    :param scale: The scale to use in the reducer. It defaults to 10 due
        to the minimum scale available in EE (Sentinel 10m)
    :type scale: int
    :param id: image property that will be the key in the result dict
    :type id: str
    :param properties: image properties that will be added to the resulting
        dict
    :type properties: list
    :param side: 'server' or 'client' side
    :type side: str
    :return: Values of all bands in the ponit
    :rtype: dict
    """
    if reducer is None:
        reducer = ee.Reducer.mean()

    if not scale:
        scale = 1
    else:
        scale = int(scale)

    if not properties:
        properties = []
    properties = ee.List(properties)

    def listval(img, it):
        theid = ee.Algorithms.String(img.get(id))
        values = img.reduceRegion(reducer, geometry, scale,
                                  maxPixels=maxPixels)
        values = ee.Dictionary(values)
        img_props = img.propertyNames()

        def add_properties(prop, ini):
            ini = ee.Dictionary(ini)
            condition = img_props.contains(prop)
            def true():
                value = img.get(prop)
                return ini.set(prop, value)
            return ee.Algorithms.If(condition, true(), ini)

        with_prop = ee.Dictionary(properties.iterate(add_properties, values))
        return ee.Dictionary(it).set(theid, with_prop)

    result = collection.iterate(listval, ee.Dictionary({}))
    result = ee.Dictionary(ee.Algorithms.If(collection.size().neq(0),
                                            result, {}))

    if side == 'server':
        return result
    elif side == 'client':
        return result.getInfo()
    else:
        raise ValueError("side parameter must be 'server' or 'client'")


def outliers(collection, bands, sigma=2, updateMask=False):
    """ Compute outliers by:

    outlier = value > mean+(sigma*stddev)
    outlier = value < mean-(sigma*stddev)

    Example (sigma = 1):
        - values = [1, 5, 6, 4, 7, 10]
        - mean = 5.5
        - std dev = 3
        - mean + (sigma*stddev) = 8.5
        - mean - (sigma*stddev) = 2.5
        - outliers = values between 2.5 and 8.5 = [1, 10]

    if `updateMask` is False return the passed collection in which each image
    have new bands (a mask) corresponding to the passed dict and a suffix '_outlier'
    else return the passed collection with the passed bands masked if are
    outliers (the outlier band is not returned).

    idea from: https://www.kdnuggets.com/2017/02/removing-outliers-standard-deviation-python.html
    """
    bands = bands or ee.Image(collection.first()).bandNames()
    bands = ee.List(bands)
    forstats = collection.select(bands)
    mean = forstats.mean()
    stddev = forstats.reduce(ee.Reducer.stdDev())
    imin = mean.subtract(stddev.multiply(sigma))
    imax = mean.add(stddev.multiply(sigma))

    def getOutlier(im, imin, imax):
        ismin = im.lt(imin)
        ismax = im.gt(imax)
        outlier = ismin.Or(ismax)
        return outlier

    def overcol(im):
        outs = getOutlier(im.select(bands), imin, imax)
        if updateMask:
            ibands = im.select(bands)
            ibands = ibands.updateMask(outs.Not())
        else:
            ibands = image_module.addSuffix(outs, '_outlier')

        return im.addBands(ibands, overwrite=True)

    return collection.map(overcol)


def data2pandas(data):
    """
    Convert data coming from tools.imagecollection.get_values to a
    pandas DataFrame

    :type data: dict
    :rtype: pandas.DataFrame
    """
    # Indices
    # header
    allbands = [val.keys() for bands, val in data.items()]
    header = []
    for bandlist in allbands:
        for band in bandlist:
            if band not in header:
                header.append(band)

    data_dict = {}
    indices = []
    for i, head in enumerate(header):
        band_data = []
        for iid, val in data.items():
            if i == 0:
                indices.append(iid)
            band_data.append(val[head])
        data_dict[head] = band_data

    df = pd.DataFrame(data=data_dict, index=indices)

    return df


def parametrizeProperty(collection, property, range_from, range_to,
                        pattern='{property}_PARAMETRIZED'):
    """ Parametrize a property

    :param collection: the ImageCollection
    :param range_from: the original property range
    :param range_to: the desired property range
    :param property: the name of the property
    :param pattern: the name of the resulting property. Wherever it says
        'property' will be replaced with the passed property.
    :return: the parsed collection in which every image has a new
        parametrized property
    """
    name = pattern.replace('{property}', property)

    original_range = range_from if isinstance(range_from, ee.List) \
        else ee.List(range_from)

    final_range = range_to if isinstance(range_to, ee.List) \
        else ee.List(range_to)

    # original min and max
    min0 = ee.Number(original_range.get(0))
    max0 = ee.Number(original_range.get(1))

    # range from min to max
    rango0 = max0.subtract(min0)

    # final min max images
    min1 = ee.Number(final_range.get(0))
    max1 = ee.Number(final_range.get(1))

    rango1 = max1.subtract(min1)

    def wrap(img):
        value = ee.Number(img.get(property))
        percent = value.subtract(min0).divide(rango0)
        final = percent.multiply(rango1).add(min1)
        return img.set(name, final)

    return collection.map(wrap)


def linearFunctionBand(collection, band, range_min=None, range_max=None,
                       mean=None, output_min=None, output_max=None,
                       name='linear_function'):
    """ Apply a linear function over the bands across every image of the
    ImageCollection using the following formula:

    - a = abs(val-mean)
    - b = output_max-output_min
    - c = abs(range_max-mean)
    - d = abs(range_min-mean)
    - e = max(c, d)

    f(x) = a*(-1)*(b/e)+output_max

    :param band: the band to process
    :param range_min: the minimum pixel value in the parsed band. If None, it
        will be computed reducing the collection
    :param range_max: the maximum pixel value in the parsed band. If None, it
        will be computed reducing the collection
    :param output_min: the minimum value that will take the resulting band.
    :param output_max: the minimum value that will take the resulting band.
    :param mean: the value on the given range that will take the `output_max`
        value
    :param name: the name of the resulting band
    :return: the parsed collection in which every image will have an extra band
        that results of applying the linear function over every pixel in the
        image
    :rtype: ee.ImageCollection
    """
    if range_min is None:
        range_min = ee.Image(collection.select(band).min()).rename('imin')
    else:
        range_min = castImage(range_min)

    if range_max is None:
        range_max = ee.Image(collection.select(band).max()).rename('imax')
    else:
        range_max = castImage(range_max)

    def to_map(img):
        result = image_module.linearFunction(img, band, range_min, range_max,
                                             mean, output_min, output_max,
                                             name)
        return img.addBands(result.rename(name))

    collection = collection.map(to_map)

    return collection


def linearFunctionProperty(collection, property, range_min=None,
                           range_max=None, mean=None, output_min=None,
                           output_max=None, name='LINEAR_FUNCTION'):
    """ Apply a linear function over the properties across every image of the
    ImageCollection using the following formula:

    - a = abs(val-mean)
    - b = output_max-output_min
    - c = abs(range_max-mean)
    - d = abs(range_min-mean)
    - e = max(c, d)

    f(x) = a*(-1)*(b/e)+output_max

    :param property: the property to process
    :param range_min: the minimum pixel value in the parsed band. If None, it
        will be computed reducing the collection
    :param range_max: the maximum pixel value in the parsed band. If None, it
        will be computed reducing the collection
    :param output_min: the minimum value that will take the resulting band.
    :param output_max: the minimum value that will take the resulting band.
    :param mean: the value on the given range that will take the `output_max`
        value
    :param name: the name of the resulting band
    :return: the parsed collection in which every image will have an extra
        property that results of applying the linear function over every pixel
        in the image
    :rtype: ee.ImageCollection
    """
    if range_min is None:
        imin = ee.Number(collection.aggregate_min(property))
    else:
        imin = ee.Number(range_min)

    if range_max is None:
        imax = ee.Number(collection.aggregate_max(property))
    else:
        imax = ee.Number(range_max)

    if mean is None:
        imean = imax
    else:
        imean = ee.Number(mean)

    if output_max is None:
        output_max = imax
    else:
        output_max = ee.Number(output_max)

    if output_min is None:
        output_min = imin
    else:
        output_min = ee.Number(output_min)

    a = imax.subtract(imean).abs()
    b = imin.subtract(imean).abs()
    t = a.max(b)

    def to_map(img):
        val = ee.Number(img.get(property))

        a = val.subtract(imean).abs().multiply(-1)
        b = output_max.subtract(output_min)
        c = b.divide(t)
        d = a.multiply(c)
        result = d.add(output_max)

        return img.set(name, result)

    collection = collection.map(to_map)

    return collection


def gaussFunctionBand(collection, band, range_min=None, range_max=None,
                      mean=0, output_min=None, output_max=1, std=None,
                      stretch=1, name='gauss'):
    """ Compute a Guass function using a specified band over an
        ImageCollection. See: https://en.wikipedia.org/wiki/Gaussian_function

    :param band: the name of the band to use
    :type band: str
    :param range_min: the minimum pixel value in the parsed band. If None, it
        will be computed
    :param range_max: the maximum pixel value in the parsed band. If None, it
        will be computed
    :param mean: the position of the center of the peak. Defaults to 0
    :type mean: int or float
    :param std: the standard deviation value. Defaults to range/4
    :type std: int or float
    :param output_max: height of the curve's peak
    :type output_max: int or float
    :param output_min: the desired minimum of the curve
    :type output_min: int or float
    :param stretch: a stretching value. As bigger as stretch
    :type stretch: int or float
    :param name: the name of the resulting band
    :return: the parsed collection in which every image will have an extra band
        that results of applying the gauss function over every pixel in the
        image
    :rtype: ee.ImageCollection
    """
    if range_min is None:
        range_min = ee.Image(collection.min())
    else:
        range_min = castImage(range_min)

    if range_max is None:
        range_max = ee.Image(collection.max())
    else:
        range_max = castImage(range_max)

    def to_map(img):

        result = image_module.gaussFunction(img, band,
                                            range_min=range_min,
                                            range_max=range_max,
                                            mean=mean, std=std,
                                            output_min=output_min,
                                            output_max=output_max,
                                            stretch=stretch,
                                            name=name)
        return img.addBands(result)

    collection = collection.map(to_map)

    return collection


def gaussFunctionProperty(collection, property, range_min=None,
                          range_max=None, mean=0, output_min=None,
                          output_max=1, std=None, stretch=1,
                          name='GAUSS'):
    """ Compute a Guass function using a specified property over an
        ImageCollection. See: https://en.wikipedia.org/wiki/Gaussian_function

    :param collection:
    :type collection: ee.ImageCollection
    :param property: the name of the property to use
    :type property: str
    :param range_min: the minimum pixel value in the parsed band. If None, it
        will be computed
    :param range_max: the maximum pixel value in the parsed band. If None, it
        will be computed
    :param mean: the position of the center of the peak. Defaults to 0
    :type mean: int or float
    :param std: the standard deviation value. Defaults to range/4
    :type std: int or float
    :param output_max: height of the curve's peak
    :type output_max: int or float
    :param output_min: the desired minimum of the curve
    :type output_min: int or float
    :param stretch: a stretching value. As bigger as stretch
    :type stretch: int or float
    :param name: the name of the resulting property
    :return: the parsed collection in which every image will have an extra
        property that results of applying the linear function over every pixel
        in the image
    :rtype: ee.ImageCollection
    """
    if range_min is None:
        range_min = ee.Number(collection.aggregate_min(property))
    else:
        range_min = ee.Number(range_min)

    if range_max is None:
        range_max = ee.Number(collection.aggregate_max(property))
    else:
        range_max = ee.Number(range_max)

    mean = ee.Number(mean)
    output_max = ee.Number(output_max)
    if std is None:
        std = range_max.subtract(range_min).divide(4)
    else:
        std = ee.Number(std)
    stretch = ee.Number(stretch)

    def to_map(img):
        def compute_gauss(value):
            a = value.subtract(mean).pow(2)
            b = std.pow(2).multiply(-2)
            c = a.divide(b).multiply(stretch)
            d = c.exp()
            return d.multiply(output_max)

        no_parametrized = compute_gauss(ee.Number(img.get(property)))

        if output_min is None:
            return img.set(name, no_parametrized)
        else:
            min_result = compute_gauss(range_min)
            max_result = compute_gauss(range_max)
            min_result_final = min_result.min(max_result)
            e = no_parametrized.subtract(min_result_final)
            f = output_max.subtract(min_result_final)
            g = output_max.subtract(output_min)
            parametrized = e.divide(f).multiply(g).add(output_min)
            return img.set(name, parametrized)

    collection = collection.map(to_map)

    return collection


def normalDistributionProperty(collection, property, mean=None, std=None,
                               name='NORMAL_DISTRIBUTION'):
    """ Compute a normal distribution using a specified property, over an
    ImageCollection. For more see:
    https://en.wikipedia.org/wiki/Normal_distribution

    :param property: the name of the property to use
    :type property: str
    :param mean: the mean value. If None it will be computed from the source.
        defaults to None.
    :type mean: float
    :param std: the standard deviation value. If None it will be computed from
        the source. Defaults to None.
    :type std: float
    """
    if mean is None:
        imean = ee.Number(collection.aggregate_mean(property))
    else:
        imean = ee.Number(mean)

    if std is None:
        istd = ee.Number(collection.aggregate_total_sd(property))
    else:
        istd = ee.Number(std)

    imax = ee.Number(1)\
             .divide(istd.multiply(ee.Number(2).multiply(math.pi).sqrt()))

    return gaussFunctionProperty(collection, property, mean=imean,
                                 output_max=imax, std=istd, name=name)


def normalDistributionBand(collection, band, mean=None, std=None,
                           name='normal_distribution'):
    """ Compute a normal distribution using a specified band, over an
    ImageCollection. For more see:
    https://en.wikipedia.org/wiki/Normal_distribution

    :param band: the name of the property to use
    :type band: str
    :param mean: the mean value. If None it will be computed from the source.
        defaults to None.
    :type mean: float
    :param std: the standard deviation value. If None it will be computed from
        the source. Defaults to None.
    :type std: float
    """
    if mean is None:
        imean = ee.Image(collection.mean())
    else:
        imean = ee.Image.constant(mean)

    if std is None:
        istd = ee.Image(collection.reduce(ee.Reducer.stdDev()))
    else:
        istd = ee.Image.constant(std)

    ipi = ee.Image.constant(math.pi)

    imax = ee.Image(1) \
             .divide(istd.multiply(ee.Image.constant(2).multiply(ipi).sqrt()))

    return gaussFunctionBand(collection, band, mean=imean,
                             output_max=imax, std=istd, name=name)


# coding=utf-8

from __future__ import print_function
from . import tools
import ee

def binary(conditions, classes, mask_name='dt_mask'):

    cond = ee.Dictionary(conditions)
    paths = ee.Dictionary(classes)

    def C(condition, bool):
        # b = ee.Number(bool)
        return ee.Image(ee.Algorithms.If(bool, ee.Image(condition),
                                         ee.Image(condition).Not()))

    # function to iterate over the path (classes)
    def overpath(key, path):
        v = ee.List(path) # the path is a list of lists
        # define an intial image = 1 with one band with the name of the class
        ini = ee.Image.constant(1).select([0], [key])

        # iterate over the path (first arg is a pair: [cond, bool])
        def toiterate(pair, init):
            init = ee.Image(init)
            pair = ee.List(pair)
            boolean = pair.get(1)
            condition_key = pair.get(0)  # could need var casting
            condition = cond.get(condition_key)
            final_condition = C(condition, boolean)
            return ee.Image(init.And(ee.Image(final_condition)))

        result = ee.Image(v.iterate(toiterate, ini))
        return result

    new_classes = ee.Dictionary(paths.map(overpath))

    # UNIFY CLASSES. example: {'snow-1':x, 'snow-2':y} into {'snow': x.and(y)}
    new_classes_list = new_classes.keys()

    def mapclasses(el):
        return ee.String(el).split('-').get(0)

    repeated = new_classes_list.map(mapclasses)

    unique = tools.ee_list.removeDuplicates(repeated)

    # CREATE INITIAL DICT
    def createinitial(baseclass, ini):
        ini = ee.Dictionary(ini)
        i = ee.Image.constant(0).select([0], [baseclass])
        return ini.set(baseclass, i)

    ini = ee.Dictionary(unique.iterate(createinitial, ee.Dictionary({})))

    def unify(key, init):
        init = ee.Dictionary(init)
        baseclass = ee.String(key).split('-').get(0)
        mask_before = ee.Image(init.get(baseclass))
        mask = new_classes.get(key)
        new_mask = mask_before.Or(mask)
        return init.set(baseclass, new_mask)

    new_classes_unique = ee.Dictionary(new_classes_list.iterate(unify, ini))

    masks = new_classes_unique.values() # list of masks

    # Return an Image with one band per option

    def tomaskimg(mask, ini):
        ini = ee.Image(ini)
        return ini.addBands(mask)

    mask_img = ee.Image(masks.slice(1).iterate(tomaskimg,
                                               ee.Image(masks.get(0))))
    # print(mask_img)

    init = ee.Image.constant(0).rename(mask_name)

    def iterate_results(mask, ini):
        ini = ee.Image(ini)
        return ini.Or(mask)

    result = masks.iterate(iterate_results, init)

    not_mask = ee.Image(result).Not()

    return mask_img.addBands(not_mask)


# coding=utf-8

import ee
from geetools import cloud_mask
from geetools.tools.image import getValue
from . import TEST_CLOUD_IMAGES

# Initialize
ee.Initialize()

p_cloud = ee.Geometry.Point([-71.44110783027168, -42.760436204035514])
p_cirrus = ee.Geometry.Point([-71.90539080580817, -43.02237648453567])

image = ee.Image(TEST_CLOUD_IMAGES['S2'])


def test_clouds():
    masked = cloud_mask.sentinel2(['cloud'])(image)
    vals = getValue(masked, p_cloud, 30)

    assert vals.get("B1").getInfo() == None


def test_cirrus():
    masked = cloud_mask.sentinel2(['cirrus'])(image)
    vals = getValue(masked, p_cirrus, 30)

    assert vals.get("B1").getInfo() == None


# coding=utf-8
""" Authentication functions """
import ee
from google.oauth2.credentials import Credentials
import json
import os
import shutil

DEFAULT_PATH = os.path.split(ee.oauth.get_credentials_path())[0]


def list_users(credential_path=DEFAULT_PATH):
    """ List local users """
    return os.listdir(credential_path)


def rename_current_user(name, credential_path=DEFAULT_PATH):
    """ Rename the current user. If you run `ee.Initialize()` after this
    you will be ask to initialize.
    """
    origin = ee.oauth.get_credentials_path()
    destination = os.path.join(credential_path, name)
    shutil.move(origin, destination)


def Initialize(filename, credential_path='default'):
    """
    Authenticate to GEE with the specified credentials

    If credential_path is set to 'defualt', it searches for the 'filename' in
    the same folder in which credentials are stored locally
    """
    if credential_path == 'default':
        credential_path = os.path.split(ee.oauth.get_credentials_path())[0]

    path = os.path.join(credential_path, filename)

    def get_credentials():
        try:
            tokens = json.load(open(path))
            refresh_token = tokens['refresh_token']
            return Credentials(
                None,
                refresh_token=refresh_token,
                token_uri=ee.oauth.TOKEN_URI,
                client_id=ee.oauth.CLIENT_ID,
                client_secret=ee.oauth.CLIENT_SECRET,
                scopes=ee.oauth.SCOPES)
        except IOError:
            raise ee.EEException(
                'Please authorize access to your Earth Engine account by '
                'running\n\nearthengine authenticate\n\nin your command line, and then '
                'retry.')

    credentials = get_credentials()
    ee.Initialize(credentials)


""" subpackage holding modules with tools """

from . import collection, date, dictionary, ee_list, featurecollection, \
              geometry, image, imagecollection, number, string, array, \
              computedobject, element


# coding=utf-8

import ee
from geetools.tools.image import getValue
ee.Initialize()


l8SR = ee.Image("LANDSAT/LC8_SR/LC82310772014043")
p_l8SR_no_cloud = ee.Geometry.Point([-66.0306, -24.9338])

def test_expressions():
    from geetools import expressions
    generator = expressions.Expression()
    exp_max = generator.max("b('B1')", "b('B2')")
    exp_min = generator.min("b('B1')", "b('B2')")

    img_max = l8SR.expression(exp_max).select([0], ["max"])
    img_min = l8SR.expression(exp_min).select([0], ["min"])

    vals_max = getValue(img_max, p_l8SR_no_cloud, 30, 'client')
    vals_min = getValue(img_min, p_l8SR_no_cloud, 30, 'client')

    assert vals_max["max"] == 580
    assert vals_min["min"] == 517